package v10.name.panitz.wip.tree;
import java.util.Set;
import java.util.HashSet;

public class CollectFreeVars extends AbstractVisitor<Set<String>>{
  Set<String> boundVars;
  Set<String> freeVars = new HashSet<String>();

  Set<String> funNames;
  Set<String> conNames;

  public CollectFreeVars(Set<String> conNames,Set<String> funNames,Set<String> boundVars){
    this.boundVars = boundVars;
    this.funNames = funNames;
    this.conNames = conNames;
  }

  public Set<String> visit(LetExpr let) throws Exception{
    Set<String> newBound=new HashSet<>();
    for (Pair<String,Tree> arg:let.args){
      if (!boundVars.contains(arg.e1)) newBound.add(arg.e1);
      boundVars.add(arg.e1);
      arg.e2.welcome(this);
    }

    let.expr.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<String> visit(CaseExpr cas) throws Exception{
    Set<String> newBound=new HashSet<>();
    if (!cas.isVarCase){
      for (String arg:cas.args){
        if (!boundVars.contains(arg)) newBound.add(arg);
      }
    }else if (!boundVars.contains(cas.name)){
      newBound.add(cas.name);
    }
    boundVars.addAll(newBound);
    cas.def.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<String> visit(ProjectionExpr match) throws Exception{
    match.expr.welcome(this);
    return freeVars;
  }

  public Set<String> visit(MatchExpr match) throws Exception{
    match.expr.welcome(this);
    for (CaseExpr cas:match.cases){
      cas.welcome(this);
    }
    return freeVars;
  }
  public Set<String> visit(LambdaExpr lambda) throws Exception{
    Set<String> newBound=new HashSet<>();
    for (String arg:lambda.args){
      if (!boundVars.contains(arg)) newBound.add(arg);
    }
    boundVars.addAll(newBound);
    lambda.def.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<String> visit(Var v) throws Exception{
    if(!boundVars.contains(v.name))
      freeVars.add(v.name);
    return freeVars;
  }
  public Set<String> visit(FunCall funCall) throws Exception{
    if (!conNames.contains(funCall.name)&&!funNames.contains(funCall.name) && !boundVars.contains(funCall.name))
      freeVars.add(funCall.name);

    for (Tree arg:funCall.args)arg.welcome(this);
    return freeVars;
  }

  public Set<String> visit(IntLit il) throws Exception{ 
    return freeVars;
  }
  public Set<String> visit(StringLit il) throws Exception{ 
    return freeVars;
  }

  public Set<String> visit(MultExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(DivExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(ModExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(AddExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(SubExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(EqExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(NeqExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(LeExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(GeExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(LtExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(GtExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(OrExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(AndExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<String> visit(IfExpr ex) throws Exception{
    ex.cond.welcome(this);
    ex.alt1.welcome(this);
    ex.alt2.welcome(this);
    return freeVars;
  }

}

