package v10.name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.io.*;
import java.util.stream.Collectors;
import static v10.name.panitz.wip.tree.ExceptionFunction.tryCatch;

public class GlobalizeCases extends Copy{
  List<FunDef> caseFuns = new LinkedList<>();
  Set<String> funNames=new HashSet<>();
  Set<String> conNames=new HashSet<>();

  public Tree visit(Program prog) throws Exception{
    for (FunDef fun:prog.funs) funNames.add(fun.name);
    for (ConDef con:prog.cons) conNames.add(con.name);

    Program p = (Program)super.visit(prog);
    p.funs.addAll(caseFuns);
    return p;
  }

  public Tree visit(CaseExpr cas) throws Exception{
    Set<String> boundVars = new HashSet<>();
    if (cas.isVarCase)boundVars.add(cas.name); else boundVars.addAll(cas.args);
    Set<String> freeVars = cas.def.welcome(new CollectFreeVars(conNames,funNames,boundVars));
    String caseName = caseFun(); 
    
    List<String> funArgs = new LinkedList<>();

    if (cas.isVarCase){
      funArgs.add(cas.name);
    }else{
      funArgs.add("this");
      funArgs.addAll(cas.args);
    }
    funArgs.addAll(freeVars);

    FunDef f = new FunDef(caseName,funArgs,cas.def.welcome(this));
    f.isSingleCalled = true;

    caseFuns.add(f);
    funNames.add(caseName);
    CaseExpr result = new CaseExpr
     (cas.name,cas.args
     ,new FunCall
       (caseName
       ,funArgs.stream()
               .map(x->new Var(x)).collect(Collectors.toList())));

    result.isVarCase = cas.isVarCase;
    return result;
  }

  public Tree visit(LetExpr let) throws Exception{
    Set<String> boundVars = new HashSet<>();
    for (Pair<String,Tree> arg:let.args) boundVars.add(arg.e1);

    Set<String> freeVars = let.expr.welcome(new CollectFreeVars(conNames,funNames,boundVars));

    String caseName = "let"+caseFun(); 
    
    List<String> funArgs = new LinkedList<>();

    for (Pair<String,Tree> arg:let.args)   funArgs.add(arg.e1);

    funArgs.addAll(freeVars);

    FunDef f = new FunDef(caseName,funArgs,let.expr.welcome(this));
    f.isSingleCalled = true;

    caseFuns.add(f);
    funNames.add(caseName);
    LetExpr result = new LetExpr
     (let.args.stream()
              .map(tryCatch(p->new Pair<>(p.e1,p.e2.welcome(this)))).collect(Collectors.toList())
     ,new FunCall
       (caseName
       ,funArgs.stream()
               .map(x->new Var(x)).collect(Collectors.toList())));

    return result;
  }


  int nr = 0;
  synchronized String caseFun(){
    return "case"+(nr++);
  }
}

