package v4.name.panitz.wip.stackmachine;
import  v4.name.panitz.wip.stackmachine.Instruction.Instr;
import java.io.*;
import static v4.name.panitz.wip.tree.GenCode.*;

public class WriteX86{
  public boolean _32 = true;

  String mov(){return _32?"movl":"movq";}
  String ax(){return _32?"%eax":"%rax";}
  String bx(){return _32?"%ebx":"%rbx";}
  String cx(){return _32?"%ecx":"%rcx";}
  String dx(){return _32?"%edx":"%rdx";}
  String sp(){return _32?"%esp":"%rsp";}
  String bp(){return _32?"%ebp":"%rbp";}
  String push(){return _32?"pushl":"pushq";}
  String pop(){return _32?"popl":"popq";}
  String mul(){return _32?"mull":"mulq";}
  String add(){return _32?"addl":"addq";}
  String sub(){return _32?"subl":"subq";}
  String div(){return _32?"divl":"divq";}
  String and(){return _32?"andl":"andq";}
  String or(){return _32?"orl":"orq";}

  String cmp(){return _32?"cmpl":"cmpq";}
  String jmp(){return "jmp";}
  String jne(){return "jne";}
  String jnl(){return "jnl";}


  public Instruction[] code;
  public WriteX86(Instruction[] code){
    this.code = code;
  }

  private void genCompareExpr(Writer out,String jmp)throws Exception{
    out.write("   "+pop()+"   "+bx()+"\n");
    out.write("   "+pop()+"   "+ax()+"\n");
    out.write("   "+cmp()+"   "+ax()+", "+bx()+"\n");

    String endLabel = label();
    String posLabel = label();

    out.write("   "+jmp+"   "+posLabel+"\n");
    out.write("   "+push()+"   $0\n");
    out.write("   "+jmp()+"   "+endLabel+"\n");
    out.write("   "+posLabel+":\n");
    out.write("   "+push()+"   $1\n");
    out.write("   "+endLabel+":\n");
  }

  public void writeAssembler(Writer out)throws Exception{
    out.write(".globl go\n");
    out.write("go:\n");
    out.write("   "+push()+"   "+bp()+"\n");
    out.write("   "+mov()+"    "+sp()+", "+bp()+"\n");

    for (Instruction in:code){
      writeAssembler(out,in);
    }

    out.write("   "+pop()+"    "+ax()+"\n");
    out.write("   "+pop()+"    "+bp()+"\n");
    out.write("   ret\n");
    out.close();
  }

  public void writeAssembler(Writer out,Instruction ins)throws Exception{
    switch (ins.instr) {
    case PushInt:
      out.write("   "+push()+"   $"+ins.a1+"\n");
      break;

    case Mult: 
      out.write("   "+pop()+"   "+bx()+"\n");
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+mul()+"   "+bx()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Add: 
      out.write("   "+pop()+"   "+bx()+"\n");
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+add()+"   "+bx()+", "+ax()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Sub: 
      out.write("   "+pop()+"   "+bx()+"\n");
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+sub()+"   "+bx()+", "+ax()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Div:
      out.write("   "+pop() +"   "+bx()+"\n");
      out.write("   "+pop() +"   "+ax()+"\n");
      out.write("   "+mov() +"   $0, "+dx()+"\n");
      out.write("   "+div()+"   "+bx()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Mod:
      out.write("   "+pop() +"   "+bx()+"\n");
      out.write("   "+pop() +"   "+ax()+"\n");
      out.write("   "+mov() +"   $0, "+dx()+"\n");
      out.write("   "+div()+"   "+bx()+"\n");
      out.write("   "+push()+"   "+dx()+"\n");
      break;

    case Eq:
      genCompareExpr(out,"je");
      break;

    case Neq:
      genCompareExpr(out,"jne");
      break;

    case Lt:
      genCompareExpr(out,"jl");
      break;

    case Le:
      genCompareExpr(out,"jle");
      break;

    case Gt:
      genCompareExpr(out,"jg");
      break;

    case Ge:
      genCompareExpr(out,"jge");
      break;

    case And: 
      out.write("   "+pop()+"   "+bx()+"\n");
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+and()+"   "+bx()+", "+ax()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Or: 
      out.write("   "+pop()+"   "+bx()+"\n");
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+or()+"   "+bx()+", "+ax()+"\n");
      out.write("   "+push()+"   "+ax()+"\n");
      break;

    case Jmp: 
      out.write("   "+jmp()+"   "+ins.label+"\n");
      break;

    case Label: 
      out.write(ins.label+":\n");
      break;

    case JmpTrue: 
      out.write("   "+pop()+"   "+ax()+"\n");
      out.write("   "+mov()+"   $1, "+bx()+"\n");
      out.write("   "+cmp()+"   "+ax()+", "+bx()+"\n");
      out.write("   je    "+ins.label+"\n");
      break;
    }
  }
}

