package v5.name.panitz.wip.tree;
import v5.name.panitz.wip.stackmachine.*;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;

public class GenCode implements Visitor<List<Instruction>>{
  List<Instruction> result = new LinkedList<>();
  public Map<String,Integer> labels = new HashMap<>();
  Map<String,Integer> varEnv = new HashMap<>();
  Map<String,FunDef> funs = new HashMap<>();
  int start=0;

  public StackMachine getStackMachine(){
    StackMachine res = new StackMachine();
    List<Instruction> instr= new LinkedList<>();
    instr.addAll(result);
    instr.add(new Instruction(Instruction.Instr.Call,"go"));
    res.code = instr.toArray(new Instruction[0]);
    res.labels = labels;
    res.ic = result.size();
    return res;
  }

  static int lbl = 0;
  public static String label() {
    lbl++;
    return "label"+lbl;
  }


  public List<Instruction> visit(Program prog) throws Exception{
    funs = prog.welcome(new CollectFunDefs());
    for (FunDef fun:prog.funs){      
      fun.welcome(this);
    }
    start = result.size();
    return result;
  }

  public List<Instruction> visit(FunDef fun)  throws Exception{
    result.add(new Instruction(Instruction.Instr.GlobLabel,fun.name));
    labels.put(fun.name,result.size());
    result.add(new Instruction(Instruction.Instr.StartCall));


    int i = fun.args.size();
    Map<String,Integer> newVarEnv = new HashMap<>();
    for (String arg:fun.args){
      newVarEnv.put(arg,i--);  
    }
    Map<String,Integer> oldVarEnv = varEnv;
    varEnv = newVarEnv;

    fun.def.welcome(this);

    varEnv = newVarEnv;
    result.add(new Instruction(Instruction.Instr.Ret,fun.args.size()));
    return result;
  }

  public List<Instruction> visit(Var v) throws Exception{
    result.add(new Instruction(Instruction.Instr.Push,varEnv.get(v.name)));
    return result;
  }

  public List<Instruction> visit(FunCall fun)  throws Exception{
    for (Tree arg:fun.args){
      arg.welcome(this);      
    }
    result.add(new Instruction(Instruction.Instr.Call,fun.name));
    return result;
  }

  public List<Instruction> visit(IntLit il) {
    result.add(new Instruction(Instruction.Instr.PushInt,il.i));
    return result;
  }

  public List<Instruction> visit(IfExpr ex) throws Exception{
    String thenLabel = label();
    String endIfLabel = label();

    ex.cond.welcome(this);
    result.add(new Instruction(Instruction.Instr.JmpTrue,thenLabel));
    ex.alt2.welcome(this);
    result.add(new Instruction(Instruction.Instr.Jmp,endIfLabel));
    labels.put(thenLabel,result.size());
    result.add(new Instruction(Instruction.Instr.Label,thenLabel));
    ex.alt1.welcome(this);
    labels.put(endIfLabel,result.size());
    result.add(new Instruction(Instruction.Instr.Label,endIfLabel));
    return result;
  }

  public void visit(OpExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
  }
  public List<Instruction> visit(MultExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Mult));
    return result;
  }

  public List<Instruction>  visit(DivExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Div));
    return result;
  }
  public List<Instruction>  visit(ModExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Mod));
    return result;
  }
  public List<Instruction>  visit(AddExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Add));
    return result;
  }
  public List<Instruction>  visit(SubExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Sub));
    return result;
  }
  public List<Instruction>  visit(EqExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Eq));
    return result;
  }
  public List<Instruction>  visit(NeqExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Neq));
    return result;
  }
  public List<Instruction>  visit(LeExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Le));
    return result;
  }
  public List<Instruction>  visit(GeExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Ge));
    return result;
  }
  public List<Instruction>  visit(LtExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Lt));
    return result;
  }
  public List<Instruction>  visit(GtExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Gt));
    return result;
  }
  public List<Instruction>  visit(OrExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Or));
    return result;
  }
  public List<Instruction>  visit(AndExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.And));
    return result;
  }
}

