package v6.name.panitz.wip.stackmachine;
import java.util.Map;
import java.util.HashMap;

public class StackMachine{
  public Instruction[] code;
  public int ic = 0;

  int[] stack = new int[2000];
  int stc = 0;
  int stb = 0;

  public Map<String,Integer> labels = new HashMap<>();

  void step(){
    Instruction instruction = code[ic];

    switch (instruction.instr) {
    case PushInt: stack[stc++] = instruction.a1;ic++;break;
    case Mult: stack[stc-2] = stack[stc-2] * stack[stc-1];stc--;ic++;break;  
    case Mod:  stack[stc-2] = stack[stc-2] % stack[stc-1];stc--;ic++;break;  
    case Div:  stack[stc-2] = stack[stc-2] / stack[stc-1];stc--;ic++;break;  
    case Add:  stack[stc-2] = stack[stc-2] + stack[stc-1];stc--;ic++;break;  
    case Sub:  stack[stc-2] = stack[stc-2] - stack[stc-1];stc--;ic++;break;  
    case Eq:   stack[stc-2] = stack[stc-2] == stack[stc-1]?1:0;stc--;ic++;break;  
    case Neq:  stack[stc-2] = stack[stc-2] != stack[stc-1]?1:0;stc--;ic++;break;  
    case Lt:   stack[stc-2] = stack[stc-2] < stack[stc-1]?1:0;stc--;ic++;break;  
    case Gt:   stack[stc-2] = stack[stc-2] > stack[stc-1]?1:0;stc--;ic++;break;  
    case Le:   stack[stc-2] = stack[stc-2] <= stack[stc-1]?1:0;stc--;ic++;break;  
    case Ge:   stack[stc-2] = stack[stc-2] >= stack[stc-1]?1:0;stc--;ic++;break;  
    case And:  stack[stc-2] = (stack[stc-2]==1 && 1==stack[stc-1])?1:0;stc--;ic++;break;  
    case Or:   stack[stc-2] = (stack[stc-2]==1 || 1==stack[stc-1])?1:0;stc--;ic++;break;  
    case Jmp:  ic = labels.get(instruction.label);break;  
    case JmpTrue: 
        if (stack[stc-1]==1){
           ic = labels.get(instruction.label);
         }else {
           ic++;
         }
         stc--;
         break;  
    case Label:  ic++;break;  
    case GlobLabel:  ic++;break;  
    case Ret: 
      int args = instruction.a1;
      ic = stack[stb-1];
      stack[stb-1-args] = stack[stc-1];   
      stc = stc-2-args;
      stb = stack[stb];

      break;  
    case Push:
      stack[stc++] = stack[stb-code[ic++].a1-1];
      break;  
    case StartCall: 
      stack[stc] = stb;
      stb = stc;
      stc++;
      ic++;
      break;  
    case Call: 
      stack[stc++] = ic+1;
      ic = labels.get(instruction.label);
      break;  

    case PushFun: 
      stack[stc++] = labels.get(instruction.label);
      ic++;
      break;

    case FunPointerCall: 
      int nextIc = ic+1;
      ic = stack[--stc];
      stack[stc++] = nextIc;
      break;

    }
  }

  String showStack(){
    StringBuffer result = new StringBuffer();
    for (int i=0;i<stc;i++){
      result.append(stack[i]+" ");
    }
    result.append("\n");
    result.append("PC: "+ic+" BS: "+stb+" SP:"+stc+"\n");
    result.append(code[ic]+"\n");
    return result.toString();
  }

  public void run(){
    while (ic<code.length){
      System.out.println(showStack());
      step();
    }
    System.out.println(stack[0]);
  }

  public static void main(String[] args){
    StackMachine interp = new StackMachine();
    interp.code = new Instruction[1];
    interp.code[0] = new Instruction(Instruction.Instr.PushInt,42);
    interp.run();
  }  

  public String toString(){
    StringBuffer result = new StringBuffer();
    int i = 0;
    for (Instruction x:code){
      result.append((i++)+" "+x+"\n");
    }
    result.append("PC:  "+ic);

    return result.toString();
  }
}

