package v6.name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.*;
import java.util.stream.Collectors;
import java.util.function.Function;
import static v6.name.panitz.wip.tree.ExceptionFunction.tryCatch;

public class GlobalizeLambda extends AbstractVisitor<Tree>{
  List<FunDef> lambdaFuns = new LinkedList<>();

  public Tree visit(Program prog) throws Exception{
    Program p = new Program(prog.funs.stream().map(tryCatch((x)->(FunDef)x.welcome(this))).collect(Collectors.toList()));
    p.funs.addAll(lambdaFuns);
    return p;
  }

  public Tree visit(Var v) throws Exception{  
     return v;
  }
  
  


  public Tree visit(FunCall f) throws Exception{
    return new FunCall(f.name, f.args.stream().map(tryCatch((x)->x.welcome(this))).collect(Collectors.toList()) );
  }

  public Tree visit(FunDef f) throws Exception{
    return new FunDef(f.name, f.args, f.def.welcome(this));
  }



  public Tree visit(IntLit il) throws Exception{
    return il;
  }

  public Tree visit(IfExpr ex) throws Exception{
    return new IfExpr(ex.cond.welcome(this),ex.alt1.welcome(this),ex.alt2.welcome(this));
  }

  public Tree visit(MultExpr ex) throws Exception{
    return new MultExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree visit(DivExpr ex) throws Exception{
    return new DivExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(ModExpr ex) throws Exception{
    return new ModExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(AddExpr ex) throws Exception{
    return new AddExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(SubExpr ex) throws Exception{
    return new SubExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(EqExpr ex) throws Exception{
    return new EqExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(NeqExpr ex) throws Exception{
    return new NeqExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(LeExpr ex) throws Exception{
    return new LeExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(GeExpr ex) throws Exception{
    return new GeExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(LtExpr ex) throws Exception{
    return new LtExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(GtExpr ex) throws Exception{
    return new GtExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(OrExpr ex) throws Exception{
    return new OrExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(AndExpr ex) throws Exception{
    return new AndExpr(ex.left.welcome(this),ex.right.welcome(this));
  }

  int nr = 0;
  String lambdaFun(){
    return "lambda"+(nr++);
  }

  public Tree visit(LambdaExpr lambda) throws Exception{
    String lambdaName = lambdaFun(); 
    FunDef f = new FunDef(lambdaName,lambda.args,lambda.def.welcome(this));
    lambdaFuns.add(f);    
    return new Var(lambdaName);
  }

}

