package v6.name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.*;

public class Interpreter extends AbstractVisitor<Data>{
  Data result = new IntVal(0);
  Map<String,Data> varEnv = new HashMap<>();
  Map<String,FunDef> funs = new HashMap<>();

  public Data visit(Program prog) throws Exception{
    funs = prog.welcome(new CollectFunDefs());
    
    return new FunCall("go",new LinkedList<>()).welcome(this);
  }

  public Data visit(Var v) throws Exception{  
    result = varEnv.get(v.name);
    if (result==null){
      result = new FunName(v.name);
    }
    return result;
  }

  public Data visit(FunCall f) throws Exception{
    FunDef fun = funs.get(f.name);
    if (fun==null){
      Data funvar = varEnv.get(f.name);
      if (funvar!=null && funvar instanceof FunName){
        fun = funs.get(((FunName)funvar).name);
      }  
    }
    Map<String,Data> oldEnv = varEnv;
    HashMap<String,Data> newEnv = new HashMap<>();
    Iterator<Tree> params = f.args.iterator();

    for (String arg:fun.args){
      Data v = params.next().welcome(this);
      System.out.println(arg+"="+v);
      newEnv.put(arg,v);
    }
    varEnv = newEnv;
    result =  fun.def.welcome(this);
    varEnv = oldEnv;
    return result;
  }

  public Data visit(IntLit il) throws Exception{
    return new IntVal(il.i);
  }

  public Data visit(IfExpr ex) throws Exception{
    Data p = ex.cond.welcome(this);
    if (((IntVal)p).i == 1) return ex.alt1.welcome(this);
    return ex.alt2.welcome(this);
  }

  public Data visit(OpExpr ex) throws Exception{
    IntVal op1 = (IntVal)ex.left.welcome(this);
    IntVal op2 = (IntVal)ex.right.welcome(this);
    return new IntVal(ex.op.applyAsInt(op1.i, op2.i));
  }
  public Data visit(MultExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data visit(DivExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(ModExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(AddExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(SubExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(EqExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(NeqExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(LeExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(GeExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(LtExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(GtExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(OrExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Data  visit(AndExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
}

